//	Distributed under the Boost Software License Version 1.0 https://www.boost.org/LICENSE_1_0.txt
//	Copyright Gero Peterhoff

#ifndef BOOST_MATH_QUADMATH_GCC_IO_HPP
#define BOOST_MATH_QUADMATH_GCC_IO_HPP

#include <array>
#include <limits>
#include <iostream>
#include <algorithm>
#include <boost/config.hpp>
#include <boost/math/quadmath/quadmath_gcc.hpp>
//#include <boost/math/quadmath/quadmath_gcc_math.hpp>
#include <boost/math/cstdfloat/cstdfloat_limits.hpp>
#include <boost/utility/string_view.hpp>
#include <boost/utility/ios.hpp>

#if !defined(BOOST_NO_CXX17_HDR_CHARCONV)
#include <charconv>
#endif



#if BOOST_GCC
namespace boost
{
namespace math
{
namespace quadmath
{
using format_type = std::array<char, 8>;

#if !defined(BOOST_NO_CXX17_HDR_CHARCONV)
inline constexpr format_type	format(const std::chars_format chf, const int)	noexcept
{
	format_type
		fmt{"%.*Q"};

	switch (chf)
	{
		case std::chars_format::scientific:	fmt[4] = 'e';	break;
		case std::chars_format::fixed:		fmt[4] = 'f';	break;
		case std::chars_format::hex:		fmt[4] = 'a';	break;
		default:							fmt[4] = 'g';	break;
	}
	return fmt;
}

inline constexpr format_type	format(const std::chars_format chf)	noexcept
{
	format_type
		fmt{"%Q"};

	switch (chf)
	{
		case std::chars_format::scientific:	fmt[2] = 'e';	break;
		case std::chars_format::fixed:		fmt[2] = 'f';	break;
		case std::chars_format::hex:		fmt[2] = 'a';	break;
		default:							fmt[2] = 'g';	break;
	}
	return fmt;
}

inline std::to_chars_result		to_chars(char* first, char* last, const __float128 value, const format_type fmt, const int precision, const bool has_precision)	noexcept
{
	const ptrdiff_t
		//	additional size for hex "0x"
		ahs	= (fmt[2]=='a' || fmt[2]=='A' || fmt[4]=='a' || fmt[4]=='A') ? 2 : 0,
		//	buffer size
		bus = (last-first) + ahs + 1;
	std::string
		//	buffer
		buf(size_t(bus), char{});
	ptrdiff_t
		len;

	if (has_precision)	len = ::quadmath_snprintf(buf.data(), buf.size(), fmt.data(), precision, value);
	else				len	= ::quadmath_snprintf(buf.data(), buf.size(), fmt.data(), value);

	std::to_chars_result
		res;

	if (BOOST_UNLIKELY(len<0 || len>=bus))
	{
		//	error
		res.ptr	= last;
		res.ec	= std::errc::value_too_large;
	}
	else
	{
		//	no error
		std::copy_n(buf.data()+ahs, len-ahs, first);
		res.ptr = first + len - ahs;
		res.ec	= std::errc{};
	}
	return res;
}

inline bool	to_format(std::string& buf, const std::chars_format fmt)	noexcept
{
	//	todo improve checks
	size_t
		bus = buf.size();

	for (size_t idx=0; idx<buf.size(); ++idx)
	{
		if (!std::isdigit(buf[0]))
			buf.erase(0, 1);
	}
	if (BOOST_LIKELY(buf.size() == bus))
	{
		for (char& val : buf)
			val = char(std::tolower(val));

		switch (fmt)
		{
			case std::chars_format::scientific:
				bus = buf.find("e");
				if (bus != buf.npos && bus < buf.size()-1)	return true;
				else										return false;
			case std::chars_format::fixed:
				return true;
			case std::chars_format::hex:
				buf.insert(0, "0x");
				return true;
			default:
				//	std::chars_format::general:
				return true;
		}
	}
	else
		return false;
}
#endif	//	BOOST_NO_CXX17_HDR_CHARCONV



inline format_type	format(const std::ios_base& stream)	noexcept
{
	//	generate format for quadmath_snprintf

	format_type
		fmt{'%'};
	size_t
		idx = 0;
	const bool
		upp = boost::is_uppercase(stream);

	auto lower_upper = [upp](const char l, const char u)	noexcept
	{
		return upp ? u : l;
	};

	if (boost::is_showpos(stream))			fmt[++idx] = '+';
	if (boost::is_showpoint(stream))		fmt[++idx] = '#';

	//	on hex no precision
	if (boost::is_floating_point_hex(stream))
	{
		fmt[++idx] = 'Q';
		fmt[++idx] = lower_upper('a', 'A');
	}
	else
	{
		fmt[++idx] = '.';
		fmt[++idx] = '*';
		fmt[++idx] = 'Q';
		++idx;
		if		(boost::is_floating_point_scientific(stream))	fmt[idx] = lower_upper('e', 'E');
		else if (boost::is_floating_point_fixed(stream))		fmt[idx] = lower_upper('f', 'F');
		else													fmt[idx] = lower_upper('g', 'G');
	}
	return fmt;
};

template <typename String>
inline void	snprintf(String& print, const std::ios_base& stream, const __float128 val)	noexcept
{
	//	generate formated print

	constexpr size_t
		//	buffer size without precision
		bus = size_t(std::numeric_limits<__float128>::max_exponent10) + 8,
		//	max precision for int
		map = size_t(std::numeric_limits<int>::max()) - bus;
	const format_type
		fmt = format(stream);
	const size_t
		//	pre must fit for int for quadmath_snprintf
		pre = std::min(boost::get_precision(stream), map);
	std::string
		buf(bus+pre, char{});
	std::streamsize
		len;

	if (boost::is_floating_point_hex(stream))	len = ::quadmath_snprintf(buf.data(), buf.size(), fmt.data(), val);
	else										len = ::quadmath_snprintf(buf.data(), buf.size(), fmt.data(), int(pre), val);

	if (BOOST_LIKELY(len > 0))	print.assign(buf.cbegin(), buf.cbegin() + len);
	else						print.clear();
};

template <typename View>
inline void	decompose(View& prefix, View& value, const View& print)	noexcept
{
	//	split print into prefix and value

	using char_type = typename View::value_type;

	constexpr std::array<char_type, 4>
		hex{'0', 'x'},
		Hex{'0', 'X'};

	if (print.starts_with('+') || print.starts_with('-'))
	{
		prefix = View(print.data(), 1);
		value  = View(print.data() + 1);
	}
	else if (print.starts_with(hex.data()) || print.starts_with(Hex.data()))
	{
		prefix = View(print.data(), 2);
		value  = View(print.data() + 2);
	}
	else
	{
		value = print;
	}
}

template <typename String, typename View>
BOOST_NOINLINE void	recompose(String& print, const View& a, const View& b, const View& c)	noexcept
{
	//	join fill, prefix and value into print

	print.resize(a.size() + b.size() + c.size());

	auto
		beg = print.begin();

	beg = std::copy(a.begin(), a.end(), beg);
	beg = std::copy(b.begin(), b.end(), beg);
		  std::copy(c.begin(), c.end(), beg);
}

template <typename String, typename Char, typename Traits>
inline void	compose(String& print, const std::basic_ios<Char, Traits>& stream, const size_t width)	noexcept
{
	//	generate formated print for stream::adjustfield

	using view_type = boost::basic_string_view<Char>;

	const String
		//	for use prefix and value as string_view need one temporary value
		//	reduce memory operations
		prn = print,
		//	make fill string
		fls = String(width-print.size(), stream.fill());
	const view_type
		fill = fls;
	view_type
		prefix,
		value;

	decompose(prefix, value, view_type(prn));
	switch (boost::get_flags(stream, std::ios_base::adjustfield))
	{
		//	right is default
		case std::ios::internal:	recompose(print, prefix, fill, value);	break;
		case std::ios::left:		recompose(print, prefix, value, fill);	break;
		default:					recompose(print, fill, prefix, value);	break;
	}
}

template <typename String>
inline String	to_string(const __float128 value)	noexcept
{
	constexpr size_t
		//	buffer size with default precision 6
		bus = size_t(std::numeric_limits<__float128>::max_exponent10) + 8 + 6;
	std::string
		buf(bus, char{});
	const std::streamsize
		len = ::quadmath_snprintf(buf.data(), buf.size(), "%.6Qg", value);
	String
		res;

	if (BOOST_LIKELY(len > 0))
		res.assign(buf.cbegin(), buf.cbegin() + len);

	return res;
}

}	//	quadmath
}	//	math
}	//	boost



namespace std
{
template <typename Char, typename Traits>
BOOST_NOINLINE basic_ostream<Char, Traits>&  operator<<(basic_ostream<Char, Traits>& stream, const __float128 val)	noexcept
{
	using sentry_type	= typename basic_ostream<Char, Traits>::sentry;
	using string_type	= basic_string<Char, Traits>;

	const sentry_type
		sentry(stream);

	if (BOOST_LIKELY(bool(sentry)))
	{
		string_type
			print;

		boost::math::quadmath::snprintf(print, stream, val);

		{
			const size_t
				width = boost::get_width(stream);

			if (BOOST_UNLIKELY(!print.empty() && print.size() < width))
				boost::math::quadmath::compose(print, stream, width);
		}

		if (BOOST_LIKELY(!print.empty()))
		{
			const streamsize
				size = streamsize(print.size());

			if (stream.rdbuf()->sputn(print.data(), size) != size)
				stream.setstate(std::ios_base::badbit);
		}
		else
			stream.setstate(std::ios_base::failbit);
	}
	else
		stream.setstate(std::ios_base::badbit);

	stream.width(0);
	return stream;
}

template <typename Char, typename Traits>
BOOST_NOINLINE basic_istream<Char, Traits>&  operator>>(basic_istream<Char, Traits>& stream, __float128& val)		noexcept
{
	using sentry_type	= typename basic_istream<Char, Traits>::sentry;
	using string_type	= basic_string<Char, Traits>;

	const sentry_type
		sentry(stream);

	if (BOOST_LIKELY(bool(sentry)))
	{
		string_type
			read;

		stream >> read;

		const string
			//	convert
			parse = string(read.cbegin(), read.cend());
		char*
			end;

		val = ::strtoflt128(parse.data(), &end);

		const ptrdiff_t
			//	length that could actually be parsed
			len_good = end - parse.data(),
			//	length that could not be parsed
			len_fail = parse.size() - len_good;

		//	putback only unparsed character
		if (BOOST_UNLIKELY(len_fail > 0))
		{
			for (ptrdiff_t idx=read.size(); idx>len_good; )
				stream.putback(read[--idx]);
		}

		//	parse failed
		if (BOOST_UNLIKELY(len_good == 0))
			stream.setstate(ios_base::failbit);
	}
	else
	{
		val = 0;
		stream.setstate(std::ios_base::badbit);
	}

	return stream;
}

inline wstring	to_wstring(const __float128 val)	noexcept	{ return boost::math::quadmath::to_string<wstring>(val);}
inline string	to_string(const __float128 val)		noexcept	{ return boost::math::quadmath::to_string<string>(val);	}


#if !defined(BOOST_NO_CXX17_HDR_CHARCONV)
inline from_chars_result	from_chars(const char* first, const char* last, __float128& value, const chars_format fmt = chars_format::general)	noexcept
{
	string
		buf(first, last);
	from_chars_result
		res;

	if (BOOST_LIKELY(boost::math::quadmath::to_format(buf, fmt)))
	{
		//	no format error
		char*
			end;
		const __float128
			val = ::strtoflt128(buf.data(), &end);
		ptrdiff_t
			len = end - buf.data();

		if (BOOST_LIKELY(len > 0))
		{
			//	no convertion error
			if (fmt == chars_format::hex)
				len -= 2;
			value = val;
			res.ptr = first + len;
			res.ec	= errc{};
		}
		else
		{
			//	convertion error
			res.ptr = first;
			res.ec	= errc::result_out_of_range;
		}
	}
	else
	{
		//	format error
		res.ptr = first;
		res.ec	= errc::invalid_argument;
	}
	return res;
}

inline to_chars_result		to_chars(char* first, char* last, const __float128 value, const chars_format fmt, const int precision)	noexcept
{
	return boost::math::quadmath::to_chars(first, last, value, boost::math::quadmath::format(fmt, precision), precision, true);
}

inline to_chars_result		to_chars(char* first, char* last, const __float128 value, const chars_format fmt)	noexcept
{
	return boost::math::quadmath::to_chars(first, last, value, boost::math::quadmath::format(fmt), 0, false);
}

inline to_chars_result		to_chars(char* first, char* last, const __float128 value)	noexcept
{
	return boost::math::quadmath::to_chars(first, last, value, boost::math::quadmath::format_type{"%Qg"}, 0, false);
}
#endif	//	BOOST_NO_CXX17_HDR_CHARCONV
}	//	std

#endif	//	BOOST_GCC
#endif	//	BOOST_MATH_QUADMATH_GCC_IO_HPP

