//	Distributed under the Boost Software License Version 1.0 https://www.boost.org/LICENSE_1_0.txt
//	Copyright Gero Peterhoff

#ifndef BOOST_MATH_QUADMATH_GCC_MATH_HPP
#define BOOST_MATH_QUADMATH_GCC_MATH_HPP

#include <cmath>
#include <limits>
#include <boost/config.hpp>
#include <boost/math/quadmath/quadmath_gcc.hpp>



#if BOOST_GCC
namespace std
{
//	basic
inline __float128 abs(__float128 x)								noexcept	{ return ::fabsq(x);			}
inline __float128 fabs(__float128 x)							noexcept	{ return ::fabsq(x);			}
inline __float128 fmod(__float128 x, __float128 y)				noexcept	{ return ::fmodq(x, y);			}
inline __float128 remainder(__float128 x, __float128 y)			noexcept	{ return ::remainderq(x, y);	}
inline __float128 remquo(__float128 x, __float128 y, int* quo)	noexcept	{ return ::remquoq(x, y, quo);	}
inline __float128 fma(__float128 x, __float128 y, __float128 z)	noexcept	{ return ::fmaq(x, y, z);		}
inline __float128 fmax(__float128 x, __float128 y)				noexcept	{ return ::fmaxq(x, y);			}
inline __float128 fmin(__float128 x, __float128 y)				noexcept	{ return ::fminq(x, y);			}
inline __float128 fdim(__float128 x, __float128 y)				noexcept	{ return ::fdimq(x, y);			}
//inline __float128 nanq(const char* arg)						noexcept	{ return ::nanq(arg);			}
inline constexpr __float128	nanq(const char*const arg)			noexcept
{
	//	::nanq not support payload
	return __builtin_nanf128(arg);
}
inline constexpr __float128	nansq(const char*const arg)			noexcept
{
	return __builtin_nansf128(arg);
}

//	exponential
inline __float128 exp(__float128 x)		noexcept	{ return ::expq(x);		}
inline __float128 exp2(__float128 x)	noexcept	{ return ::exp2q(x);	}
inline __float128 expm1(__float128 x)	noexcept	{ return ::expm1q(x);	}

inline __float128 log(__float128 x)		noexcept	{ return ::logq(x);	}
inline __float128 log10(__float128 x)	noexcept	{ return ::log10q(x);	}
inline __float128 log2(__float128 x)	noexcept	{ return ::log2q(x);	}
inline __float128 log1p(__float128 x)	noexcept	{ return ::log1pq(x);	}

//	power
inline __float128 pow(__float128 b, __float128 e)	noexcept	{ return ::powq(b, e);		}
inline __float128 sqrt(__float128 x)				noexcept	{ return ::sqrtq(x);		}
inline __float128 cbrt(__float128 x)				noexcept	{ return ::cbrtq(x);		}
inline __float128 hypot(__float128 x, __float128 y)	noexcept	{ return ::hypotq(x, y);	}
inline __float128 hypot(__float128 x, __float128 y, __float128 z)	noexcept	{ return ::sqrtq(x*x + y*y + z*z);	}

//	trigonometric
inline __float128 sin(__float128 x)		noexcept	{ return ::sinq(x);		}
inline __float128 cos(__float128 x)		noexcept	{ return ::cosq(x);		}
inline __float128 tan(__float128 x)		noexcept	{ return ::tanq(x);		}

inline __float128 asin(__float128 x)	noexcept	{ return ::asinq(x);	}
inline __float128 acos(__float128 x)	noexcept	{ return ::acosq(x);	}
inline __float128 atan(__float128 x)	noexcept	{ return ::atanq(x);	}

inline __float128 atan2(__float128 y, __float128 x)				noexcept	{ return ::atan2q(y, x);		}
inline void sincos(__float128 x, __float128* s, __float128* c)	noexcept	{ return ::sincosq(x, s, c);	}

//	hyperbolic
inline __float128 sinh(__float128 x)	noexcept	{ return ::sinhq(x);	}
inline __float128 cosh(__float128 x)	noexcept	{ return ::coshq(x);	}
inline __float128 tanh(__float128 x)	noexcept	{ return ::tanhq(x);	}

inline __float128 asinh(__float128 x)	noexcept	{ return ::asinhq(x);	}
inline __float128 acosh(__float128 x)	noexcept	{ return ::acoshq(x);	}
inline __float128 atanh(__float128 x)	noexcept	{ return ::atanhq(x);	}

//	error and gamma
inline __float128 erf(__float128 x)		noexcept	{ return ::erfq(x);		}
inline __float128 erfc(__float128 x)	noexcept	{ return ::erfcq(x);	}
inline __float128 tgamma(__float128 x)	noexcept	{ return ::tgammaq(x);	}
inline __float128 lgamma(__float128 x)	noexcept	{ return ::lgammaq(x);	}

//	round
inline __float128 ceil(__float128 x)		noexcept	{ return ::ceilq(x);		}
inline __float128 floor(__float128 x)		noexcept	{ return ::floorq(x);		}
inline __float128 trunc(__float128 x)		noexcept	{ return ::truncq(x);		}
inline __float128 round(__float128 x)		noexcept	{ return ::roundq(x);		}
inline long int	  lround(__float128 x)		noexcept	{ return ::lroundq(x);		}
inline long long int llround(__float128 x)	noexcept	{ return ::llroundq(x);		}
inline __float128 nearbyint(__float128 x)	noexcept	{ return ::nearbyintq(x);	}
inline __float128 rint(__float128 x)		noexcept	{ return ::rintq(x);		}
inline long int	  lrint(__float128 x)		noexcept	{ return ::lrintq(x);		}
inline long long int llrint(__float128 x)	noexcept	{ return ::llrintq(x);		}

//	manipulation
inline __float128 frexp(__float128 x, int* exp)			noexcept	{ return ::frexpq(x, exp);		}
inline __float128 ldexp(__float128 x, int exp)			noexcept	{ return ::ldexpq(x, exp);		}
inline __float128 modf(__float128 x, __float128* iptr)	noexcept	{ return ::modfq(x, iptr);		}
inline __float128 scalbn(__float128 x, int exp)			noexcept	{ return ::scalbnq(x, exp);		}
inline __float128 scalbln(__float128 x, long int exp)	noexcept	{ return ::scalblnq(x, exp);	}
inline int		  ilogb(__float128 x)					noexcept	{ return ::ilogbq(x);			}
inline __float128 logb(__float128 x)					noexcept	{ return ::logbq(x);			}
inline __float128 nextafter(__float128 x, __float128 y)	noexcept	{ return ::nextafterq(x, y);	}
inline __float128 nexttoward(__float128 x, __float128 y)noexcept	{ return -::nextafterq(-x, -y);	}
inline __float128 copysign(__float128 x, __float128 y)	noexcept	{ return ::copysignq(x, y);		}

//	classification and comparison
inline int	fpclassify(__float128 x)	noexcept
{
    	 if (::isnanq(x))				return FP_NAN;
    else if (::isinfq(x))				return FP_INFINITE;
    else if (x == 0)					return FP_ZERO;
    else if (::fabsq(x) < FLT128_MIN)	return FP_SUBNORMAL;
    else								return FP_NORMAL;
}

inline bool signbit(__float128 x)		noexcept	{ return ::signbitq(x);	}
inline bool isfinite(__float128 x)		noexcept	{ return ::finiteq(x);	}
inline bool isinf(__float128 x)			noexcept	{ return ::isinfq(x);	}
inline bool isnan(__float128 x)			noexcept	{ return ::isnanq(x);	}
inline bool	isnormal(__float128 x)		noexcept
{
	return fpclassify(x) == FP_NORMAL;
}
inline bool isunordered(__float128 x, __float128 y)	noexcept
{
	return ::isnanq(x) || ::isnanq(y);
}
inline bool isgreater(__float128 x, __float128 y)	noexcept
{
	if (BOOST_UNLIKELY(isunordered(x, y)))	return false;
	else									return x > y;
}
inline bool isgreaterequal(__float128 x, __float128 y)	noexcept
{
	if (BOOST_UNLIKELY(isunordered(x, y)))	return false;
	else									return x >= y;
}
inline bool isless(__float128 x, __float128 y)	noexcept
{
	if (BOOST_UNLIKELY(isunordered(x, y)))	return false;
	else									return x < y;
}
inline bool islessequal(__float128 x, __float128 y)	noexcept
{
	if (BOOST_UNLIKELY(isunordered(x, y)))	return false;
	else									return x <= y;
}
inline bool islessgreater(__float128 x, __float128 y)	noexcept
{
	if (BOOST_UNLIKELY(isunordered(x, y)))	return false;
	else									return (x < y) || (x > y);
}

//	more
inline __float128 j0(__float128 x)			noexcept	{ return ::j0q(x);		}
inline __float128 j1(__float128 x)			noexcept	{ return ::j1q(x);		}
inline __float128 jn(int x, __float128 y)	noexcept	{ return ::jnq(x, y);	}

inline __float128 y0(__float128 x)			noexcept	{ return ::y0q(x);		}
inline __float128 y1(__float128 x)			noexcept	{ return ::y1q(x);		}
inline __float128 yn(int x, __float128 y)	noexcept	{ return ::ynq(x, y);	}



template <> struct numeric_limits<__float128>
{
static constexpr bool				is_specialized		= true;
static constexpr __float128			min()				noexcept	{ return FLT128_MIN;	}
static constexpr __float128			max()				noexcept	{ return FLT128_MAX;	}
static constexpr __float128			lowest()			noexcept	{ return -FLT128_MAX;	}
static constexpr int				digits				= FLT128_MANT_DIG;
static constexpr int				digits10			= FLT128_DIG;
static constexpr int				max_digits10		= 36;
static constexpr bool				is_signed			= true;
static constexpr bool				is_integer			= false;
static constexpr bool				is_exact			= false;
static constexpr int				radix				= 2;
static constexpr __float128			epsilon()			noexcept	{ return FLT128_EPSILON;	}
static constexpr __float128			round_error()		noexcept	{ return 0.5Q;				}
static constexpr int				min_exponent		= FLT128_MIN_EXP;
static constexpr int				min_exponent10		= FLT128_MIN_10_EXP;
static constexpr int				max_exponent		= FLT128_MAX_EXP;
static constexpr int				max_exponent10		= FLT128_MAX_10_EXP;
static constexpr bool				has_infinity		= true;
static constexpr bool				has_quiet_NaN		= true;
static constexpr bool				has_signaling_NaN	= false;//bool(issignalingq(0.0Q));
static constexpr float_denorm_style	has_denorm			= denorm_present;
static constexpr bool				has_denorm_loss		= false;
static constexpr __float128			infinity()			noexcept	{ return HUGE_VALQ;			}
static constexpr __float128			quiet_NaN()			noexcept	{ return std::nanq("");		}
static constexpr __float128			signaling_NaN()		noexcept	{ return std::nansq("");	}
static constexpr __float128			denorm_min()		noexcept	{ return FLT128_DENORM_MIN;	}
static constexpr bool				is_iec559			= true;
static constexpr bool				is_bounded			= true;
static constexpr bool				is_modulo			= false;
static constexpr bool				traps				= false;
static constexpr bool				tinyness_before		= false;
static constexpr float_round_style	round_style			= round_to_nearest;
};
}	//	std

#endif	//	BOOST_GCC
#endif	//	BOOST_MATH_QUADMATH_GCC_MATH_HPP
